/**
	Web Analyzer Crawler module

	Maintains the web crawling algorithm, process downloaded pages
	Jan Raszyk, 2007
*/

#include <queue>
#include <vector>
#include <string>
#include <set>
#include <fstream>
#include <sstream>
#include <map>
#include <winsock2.h>
#include "html_parse.h"
#include "crawler.h"
#include "tcpclient.h"
#include "downloader.h"
#include "common.h"
#include "sitestats.h"

using namespace std;

namespace wbn {

class CrawlSettings
{
public:
    // width of the crawling (how many links from each page we should crawl .. 0 = all)
    unsigned            crawl_width;
    // how deep (how many levels from the start page) should the crawling go
    unsigned            crawl_depth;
    // number of simultaneous streams
    unsigned            simult_streams;
	// maximum number of examined pages
	unsigned			max_pages;
    std::vector<std::string>
        patterns;
};

Crawler::Crawler() {
    _settings = new CrawlSettings;
	stats = new SiteStats;
    fpaths_inited = false;
	_state = CRAWS_IDLE;
	force_break = false;
	cssdump.open("cssdump.css", ios_base::app);
}

Crawler::~Crawler() {
    delete _settings, stats;
	cssdump.close();
}

void Crawler::AddStartUrl(CrawlURL cu) { _queued.push_back(cu); }
void Crawler::AddUrlPattern(string url) { _settings->patterns.push_back(url); }
void Crawler::SetDepth(int d) { _settings->crawl_depth = d; }
void Crawler::SetWidth(int w) { _settings->crawl_width = w; }
void Crawler::SetSimuS(int s) { _settings->simult_streams = s; }
void Crawler::SetMaxPg(int p) { _settings->max_pages = p; }
crawler_state_t		Crawler::GetState() const { return _state; }
std::string			Crawler::GetLastUrl() const
{
	return _lasturl;
}
CrawlerQueueSizes	Crawler::GetSizes() const
{
	CrawlerQueueSizes t;
	t.pending = (unsigned int) _processing.size();
	t.queued = _queued.size();
	t.visited = stats->total_pages;
	t.errorneous = _errorneous.size();
	return t;
}
void Crawler::ForceBreak (void) { this->force_break = true; }

/* thrash*
void CrawlURLs::Push(CrawlUrl cu) { _urls.push(cu); }
CrawlUrl CrawlURLs::Pop() { CrawlUrl cu = _urls.front(); _urls.pop(); return cu; }

bool VisitedURLs::Add(string url) { return _set.insert(url).second; }
*/

std::ostream & operator << (std::ostream & o, const Crawler & d)
{

    o << "opt depth " << d._settings->crawl_depth << endl;
    o << "opt width " << d._settings->crawl_width << endl;
    o << "opt simultd " << d._settings->simult_streams << endl;
	o << "opt maxpages " << d._settings->max_pages << endl;

    /* stats
    o << d._stats;
    */

    for (int i = 0; i < (int) d._settings->patterns.size(); i++)
    {
        o << "opt addmask " << d._settings->patterns[i] << endl;
    }

    for (urldlqueue::const_iterator j = d._processing.begin(); j != d._processing.end(); j++)
    {
            o << "url " << (*j)->GetDepth() << " " << (*j)->url << endl;
    }

    for (urlqueue::const_iterator cr = d._queued.begin();
        cr != d._queued.end(); cr++)
    {
        o << "url " << cr->depth << " " << cr->url << endl;
    }

    for (urlset::const_iterator vit = d._visited.begin(); vit != d._visited.end(); vit++)
    {
        o << "visited " << *vit << endl;
    }

	return d.stats->ProjectDump(o);
}

std::istream & operator >> (std::istream & i, Crawler & d)
{
    string s, s1, s2, s3;

    while (!i.eof())
    {
        getline(i, s);
        if (s.empty()) continue;

        istringstream is(s);
        is >> s1;
        if (!s1.compare("opt"))
        {
            is >> s2 >> s3;
            if      (!s2.compare("depth"))      d.SetDepth(atoi(s3.c_str()));
            else if (!s2.compare("width"))      d.SetWidth(atoi(s3.c_str()));
            else if (!s2.compare("simultd"))    d.SetSimuS(atoi(s3.c_str()));
			else if (!s2.compare("maxpages"))   d.SetMaxPg(atoi(s3.c_str()));
            else if (!s2.compare("addmask"))    d.AddUrlPattern(s3);
        }
        else if (!s1.compare("url"))
        {
            is >> s2 >> s3;
            CrawlURL cu;
            cu.depth = atoi(s2.c_str());
            cu.url = s3;
            d.AddStartUrl(cu);
        }
        else if (!s1.compare("visited"))
        {
            is >> s2;
            d._visited.insert(s2);
        }
		else if (!s1.compare("statsstart"))
		{
			i >> *d.stats;
		}
    }

    return i;
}

void Crawler::DumpData(void) const
{
    ofstream fd;

    if (!fpaths_inited) return;

    fd.open(f_data.c_str());
    if (fd.good())
        fd << *this;
    fd.close();
}

void Crawler::DumpResult(const string & s) 
{
	ofstream fd;

	fd.open(s.c_str());
	if (!fd.good()) 
	{
		fd.close();
		return;
	}
	fd << *this->stats;
	fd.close();
}

void Crawler::InitProcessing(void)
{
    _processing.clear();
}

bool Crawler::StartProjectFile(string path)
{
    f_data = path;

    ofstream fd;
    fd.open(path.c_str(), ios_base::out);
    fpaths_inited = fd.good();
    fd.close();
    
	if (!fpaths_inited) return false;

	_state = CRAWS_CRAWLING;

    DumpData();

    InitProcessing();

    return true;
}

bool Crawler::LoadFromFile(string path)
{
	ifstream fd;
	fd.open(path.c_str(), ios_base::in);
	if (!fd.good()) return false;
	fpaths_inited = true;
	f_data = path;
	fd >> *this;
	fd.close();

	DumpData();
	InitProcessing();

	return true;
}

void Crawler::AddStats(const HtmlPageGetter* p, bool has_toplevel_link)
{
	if (!p->dom)
	{
		_errorneous.insert(p->url);
	}
	PageStats ps(*(p->dom));
	ps.homepagelink = has_toplevel_link;
	*stats += ps;
}

void Crawler::QueueLinks(const HtmlPageGetter* p, bool & has_toplevel_link)
{	
	int i, c;
	unsigned a = 0;
	CrawlURL cu;
	vector<string> links;
	string fullurl;

	// check the depth
	if (p->GetDepth() >= (int) _settings->crawl_depth)
		return;

	c = p->GetLinks(links);
	cu.depth = p->GetDepth() + 1;
	has_toplevel_link = false;

	for (i = 0; i < c; i++)
	{	
		fullurl = GetAbsoluteURL(p->url, links[i]);
		if (!has_toplevel_link && IsLinkToToplevel(fullurl, p->url))
			has_toplevel_link = true;
		
		if (!WCVectorMatch(_settings->patterns, fullurl))
			continue;

		if (Visited(fullurl))
			continue;

		a++;
		cu.url = fullurl;
		_queued.push_back(cu);
		if (_settings->crawl_width && a >= _settings->crawl_width)
			break;
		
		if (_settings->max_pages <= stats->total_pages)
			break;
	}
}

void Crawler::FinishedPage(const HtmlPageGetter* p)
{
	bool has_toplevel_link;
	QueueLinks(p, has_toplevel_link);
	AddStats(p, has_toplevel_link);
	cssdump << p->csstext << endl;
	cssdump.flush();
}

// removes errorneous and finished downloads
void Crawler::ClearFinished(void)
{
    bool again = true;
    
    while (again)
    {
        again = false;

        for (urldlqueue::iterator qi = _processing.begin(); qi != _processing.end(); qi++)
        {
            if ((*qi)->Finished())
            {
                again = true;
				_visited.insert((*qi)->url);
				// each 100 pages dump data
				if (_visited.size() % 100 == 0) this->DumpData();
				
				if (stats->total_pages < _settings->max_pages)
				{
					if ((*qi)->GetState() == LS_READY)
						FinishedPage(*qi);
					else
						_errorneous.insert((*qi)->url);
				}

				delete (*qi);
                _processing.erase(qi); break; // iterators invalidated!
            }
        } // for

    } // while
}

// fill free slots with queued pages
void Crawler::ProcessQueued(void)
{
    int freeslots = _settings->simult_streams - _processing.size();
	int addednew = 0;

    if (freeslots <= 0) return;

	while (addednew < freeslots && !_queued.empty())
	{
        CrawlURL cu = _queued.front();
        _queued.pop_front();

		if (Visited(cu.url))
			continue;

		_visited.insert(cu.url);

		if (_settings->max_pages <= stats->total_pages)
			continue;

        HtmlPageGetter *pg = new HtmlPageGetter;
        _processing.push_back(pg);
		addednew++;
        
		this->_lasturl = cu.url;
        pg->StartDownload(cu.url, cu.depth);
    }
}

// clears finished downloads, adds queued ones
// returns true if there's nothing more to download
bool Crawler::Update(void)
{
    bool done;
	
	ClearFinished();
	if (!force_break)
		ProcessQueued();
	
	done = (this->_queued.empty() && this->_processing.empty()) 
		|| 
		(force_break && this->_processing.empty());
	if (done) _state = CRAWS_DONE;
	if (done) DumpData();

	return done;
}

bool Crawler::Visited(const string & s)
{
	return (_visited.find(s) != _visited.end());
}
}; // namespace wbn

