/**
	Web Analyzer Crawler module

	Maintains the web crawling algorithm, process downloaded pages
	Jan Raszyk, 2007
*/

namespace wbn {

struct CrawlURL
{
    int depth;
    std::string url;
};

class CrawlSettings;
class CrawlStats;
class CrawlURLs;
struct CrawlURL;
class HtmlPageGetter;
class SiteStats;

typedef std::deque<CrawlURL>            urlqueue;
typedef std::deque<HtmlPageGetter *>    urldlqueue;
typedef std::set<std::string>           urlset;

// state of the crawler object
typedef enum { CRAWS_IDLE, CRAWS_CRAWLING, CRAWS_DONE } crawler_state_t;

// sizes of three main queues
struct CrawlerQueueSizes { unsigned pending, queued, visited, errorneous; };

// this structure represents main application code
// it is responsible for crawling algorithm and maintaining the page queues
class Crawler
{
private:
    CrawlSettings   *_settings;					// user settings
	SiteStats		*stats;						// statistics gained from crawled pages
	urlqueue        _queued;					// pages we have to visit in future
    urldlqueue      _processing;				// pages we are currently examining
    urlset          _visited;					// pages we have already visited
	urlset			_errorneous;				// pages where download finished with an error
    std::string     f_data;						// path to the project file associated with current instance
    bool            fpaths_inited;				// is the path to the project file correct?
	bool			Visited(const std::string &);	// tells if the given URL is present in _visited
	crawler_state_t _state;						// current crawling status
	std::string		_lasturl;					// last visited page
	void			AddStats(const HtmlPageGetter* p, bool);		// adds info about the given page to the total stats
	void			QueueLinks(const HtmlPageGetter* p, bool &);	// adds links from given page into the queue
    void            FinishedPage(const HtmlPageGetter* p);	// calls tasks when a download of a page has finished
	bool			force_break;				// when set to true, no more pages will be added to be processed
	std::ofstream	cssdump;

public:
    Crawler();
    ~Crawler();

    // adds one starting page
    void                AddStartUrl(CrawlURL);		// adds given page to queued
    // adds wildcard pattern that has to be matched
    void                AddUrlPattern(std::string);	// adds given patter to user settings
    void                SetWidth(int w);			// sets crawling width (how many links from one page we are allowed to follow)
    void                SetDepth(int d);			// sets crawling depth (how deep can we crawl from page to page)
    void                SetSimuS(int s);			// sets max connections limit
	void				SetMaxPg(int p);			// set limit for total amount of visited pages
	crawler_state_t		GetState() const;			// tells the current state of crawling
	CrawlerQueueSizes	GetSizes() const;			// tells sizes of queued, visited, processing sets
	std::string			GetLastUrl() const;			// returns URL of last visited page

    bool                StartProjectFile(std::string path);	// creates a file and dumps crawling settings data into it
	bool				LoadFromFile(std::string path);	// loads project file and initializes structures
    void                DumpData(void) const;		// dumps data into project file
    bool                Update(void);				// maintains queues, main crawling functions
    void                InitProcessing(void);		// clears structures
    void                ProcessQueued(void);		// moves queued pages from queued into processing queue
    void                ClearFinished(void);		// moves downloaded pages from processing into visited queue
	void				ForceBreak(void);			// sets force_break flag to true, causes downloads to stop soon
	void				DumpResult(const std::string &);	// writes HTML formatted statistics into given file

	// for project file writing
    friend std::ostream & operator << (std::ostream &, const Crawler &);
	// for project file reading
    friend std::istream & operator >> (std::istream &, Crawler &);
}; // class Crawler

}; // namespace wbn
