/**

	Web Analyzer HTML Parser module

	Goes through the HTML input and builds Document Object Model structure
	Jan Raszyk, 2007

*/

#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <iostream>
#include "html_parse.h"

using namespace std;

#define ISWSPACE(x) ((x) == ' ' || (x) == '\t' || (x) == '\n' || (x) == '\r')

namespace wbn {

HTMLAttr::HTMLAttr(const char * _name, const char * _value) : name(_name), val(_value) {}

static bool Follows(const char *s, const char *t)
{
    while (*s && *t && *s == *t) { s++; t++; }

    return *t == '\0';
}

static void SkipWChars(const char **c)
{
    while (**c && ISWSPACE(**c))
        (*c)++;
}

void HTML_DOMTree::SkipTillElStart()
{
	const char **c = &text;
    while (**c && (**c != '<'))
        (*c)++;

    if (**c == '<')
        (*c)++;
}

void HTML_DOMTreeCSSDump::SkipTillElStart()
{
	while (*text && (*text != '<'))
	{
		if (in_css_section)
			cssexportstring += *text++;
		else
			text++;
	}

	if (*text == '<')
		text++;
}

void HTML_DOMTree::SkipCommentContent()
{
    while (*text && !Follows(text, "-->")) text++;

    if (Follows(text, "-->"))
        text += 3;
}

void HTML_DOMTreeCSSDump::SkipCommentContent()
{
	while (*text && !Follows(text, "-->")) {
		if (in_css_section)
			cssexportstring += *text++;
		else
			text++;
	}

    if (Follows(text, "-->"))
        text += 3;
}

void HTML_DOMTree::SkipCData(void)
{
    while (*text && !Follows(text, "]]>")) text++;

    if (Follows(text, "]]>"))
        text += 3;
}

void HTML_DOMTreeCSSDump::SkipCData(void)
{
    while (*text && !Follows(text, "]]>")) 
	{
		if (in_css_section)
			cssexportstring += *text++;
		else
			text++;
	}

    if (Follows(text, "]]>"))
        text += 3;
}

static string ReadDoctype(const char **c)
{
    string ret;
	string::size_type q1, q2;

    while (**c && **c != '>')
    {
        ret += **c;
        (*c)++;
    }

    if (**c == '>') 
        (*c)++;

	
	q1 = ret.find("\"");
	if (q1 != string::npos)
	{
		q2 = ret.find("\"", q1+1);
		if (q2 != string::npos)
		{
			return ret.substr(q1+1, q2-q1-1);
		}
	}

    return ret;
}

static void SkipTillElementOver(const char **c)
{
    while (**c && **c != '>')
    {
        (*c)++;
    }

    if (**c == '>')
        (*c)++;
}

static HTMLAttr *GetAttribute(const char **c)
{
    SkipWChars(c);
    static string name, value;

	name = ""; value = "";

    if (**c == '<') { (*c)--; return 0; }
    if (**c == '>') { return 0; }
    
    // name
    while (**c)
    {
        if (**c == '=') { (*c)++; break; }
        if (ISWSPACE(**c)) 
        {   // spaces
            SkipWChars(c);
            if (**c != '=') 
				return (name.empty() ? 0 : new HTMLAttr(name.c_str(), ""));
        }
        if (**c == '/' || **c == '<' || **c == '>') { 
			(*c)++;
			return (name.empty() ? 0 : new HTMLAttr(name.c_str(), ""));
		}
        name += **c;
        (*c)++;
    }

    // value
    while (**c)
    {
        SkipWChars(c);
        if (**c == '\"' || **c == '\'')
        {
            char b = **c;
            while (**c)
            {
                (*c)++;
                if (**c == b) {
					(*c)++;
					return (name.empty() ? 0 : new HTMLAttr(name.c_str(), value.c_str()));
				}
                value += **c;
            }
        }
        if (**c == '<' || **c == '>')
        {
            if (**c == '>') (*c)++;
            return name.empty() ? 0 : new HTMLAttr(name.c_str(), "");
        }
        value += **c;
        if (!**c) break;
        if (ISWSPACE(**c) || **c == '<' || **c == '>')
        {
            if (**c != '<') (*c)++;
            return name.empty() ? 0 : new HTMLAttr(name.c_str(), value.c_str());
        }

        value += **c;
        (*c)++;
    }

    return 0;
}

void HTML_DOMTree::ElementFound(const HTMLElement & e) {}
void HTML_DOMTree::ElementOver() {}

void HTML_DOMTreeCSSDump::ElementFound(const HTMLElement & e) {
	if (!e.name.compare("style")) in_css_section = true;
}
void HTML_DOMTreeCSSDump::ElementOver() {
	in_css_section = false;
}

HTMLElement* HTML_DOMTree::GetElement()
{
    HTMLElement *he = new HTMLElement;
    HTMLAttr *ap;

    while (*text && !ISWSPACE(*text) && *text != '>' && *text != '<')
    {
        he->name += *text;
        text++;
    }
    
	if (*text == '<') { delete he; return 0; }
	if (*text == '\0') { delete he; return 0; }
    
    while (ap = GetAttribute(&text))
    {
        // transform(ap->name.begin(), ap->name.end(), ap->name.begin(), tolower);
        he->attrs[ap->name] = ap->val;
        delete ap;
    }

    // transform(he.name.begin(), he.name.end(), he.name.begin(), tolower);

	ElementFound(*he);
    return he;
}

HTML_DOMTree::HTML_DOMTree(const char *html) 
: text(html) {}

HTML_DOMTreeCSSDump::HTML_DOMTreeCSSDump(const char *html, std::string & o)
: cssexportstring(o), HTML_DOMTree(html), in_css_section(false) {}

void HTML_DOMTree::Parse()
{
	HTMLElement *ep = 0;
    xmldecl = false;
	
	// debugging: create local copy
	// string buf = text;
	// text = buf.c_str();

    while (*text)
    {
        SkipTillElStart();
        if (!*text) break;

        if (Follows(text, "!--"))
        {
			text += 3;
            SkipCommentContent();
            continue;
        }
        if (Follows(text, "!DOCTYPE"))
        {
            doctype = ReadDoctype(&text);
            continue;
        }
        if (Follows(text, "?xml"))
        {
            xmldecl = true;
            SkipTillElementOver(&text);
            continue;
        }
        if (Follows(text, "!["))
        {
			text += 2;
            SkipCData();
            continue;
        }
        if (Follows(text, "/"))
        {
			ElementOver();
            SkipTillElementOver(&text);
            continue;
        }

        ep = GetElement();

        if (ep)
        {
            elems.push_back(ep);
        }
    }
}

ostream & operator << (ostream & o, const HTML_DOMTree & t)
{
    vector<HTMLElement *>::const_iterator ei;
    map<string, string>::const_iterator ai;
    if (t.xmldecl) o << "<?xml ... ?>" << endl;
    o << "<!DOCTYPE " << t.doctype << endl;
    
    for (ei = t.elems.begin(); ei != t.elems.end(); ei++)
    {
        o << "<" << (*ei)->name;
        for (ai = (*ei)->attrs.begin(); ai != (*ei)->attrs.end(); ai++)
        {
            o << endl << "  " << ai->first << "=\"" << ai->second << "\"";
        }
        o << ">" << endl;
    }

    return o;
}

HTML_DOMTree::~HTML_DOMTree()
{
	vector<HTMLElement *>::const_iterator ei;
	for (ei = elems.begin(); ei != elems.end(); ei++)
	{
		delete *ei;
	}
}


};
