/**
	
	Site stats module

	Builds a set of information about page from given Document Object Model (DOM) object
	author: Jan Raszyk
	May 2007

*/

#include <string>
#include <map>
#include <vector>
#include "html_parse.h"
#include "sitestats.h"
#include "common.h"

using namespace std;

namespace wbn {

#define FORMATTED_ACCEPTABLE 0.2

static int GetElementsCount(const HTML_DOMTree & dom, const char *elname)
{
	int ret = 0;
	string name;
	elementslist::const_iterator ei;
	
	for (ei = dom.elems.begin(); ei != dom.elems.end(); ei++)
	{
		if (!(*ei)->name.compare(elname))
		{
			ret++;
		}
	}

	return ret;
}

static int GetElementsWithAttrCount(const HTML_DOMTree & dom, const char *element, const char *attr, const char *valmask)
{
	int ret = 0;
	elementslist::const_iterator ei;

	for (ei = dom.elems.begin(); ei != dom.elems.end(); ei++)
	{
		if (!(*ei)->name.compare(element))
		{
			if (WildCardMatch(valmask, (*ei)->attrs[attr].c_str()))
			{
				ret++;
			}
		}
	}

	return ret;
}

static int GetElementsFromSetCount(const HTML_DOMTree & dom, const char **s, size_t s_size)
{
	unsigned i, ret = 0;
	for (i = 0; i < s_size; i++)
	{
		ret += GetElementsCount(dom, s[i]);
	}
	return ret;
}

static const char * formating_elements[] =
{ "b", "i", "u", "blink", "center", "font", "big", "small", "s", "basefont" };

static const char * structural_elements[] = 
{ "ol", "ul", "dl", "dt", "dd", "abbr", "address", "blockquote", "q", "caption", "cite", "code", "del", "dfn",
"em", "strong", "h1", "h2", "h3", "h4", "h5", "h6", "ins", "kbd", "samp", "var" };

PageStats::PageStats(const HTML_DOMTree & dom)
{
	this->doctype = dom.doctype;
	this->xml_decl = dom.xmldecl;
	this->format_els = GetElementsFromSetCount(dom, formating_elements, sizeof(formating_elements)/sizeof(char*));
	
	this->semantic_els = GetElementsFromSetCount(dom, structural_elements, sizeof(structural_elements)/sizeof(char*));
	
	this->image_types.jpg_images = GetElementsWithAttrCount(dom, "img", "src", "*jpg");
	this->image_types.gif_images = GetElementsWithAttrCount(dom, "img", "src", "*gif");
	this->image_types.png_images = GetElementsWithAttrCount(dom, "img", "src", "*png");
	this->image_types.other_images = GetElementsCount(dom, "img") - (
		image_types.gif_images + image_types.jpg_images + image_types.png_images);
	
	this->xml_feed = false;
	if (GetElementsWithAttrCount(dom, "link", "type", "*rss*"))
		this->xml_feed = true;

	this->javascript = false;

	if (GetElementsWithAttrCount(dom, "script", "type", "*javascript*"))
		this->javascript = true;

	if (!javascript && GetElementsWithAttrCount(dom, "link", "type", "*javascript*"))
		this->javascript = true;


	this->stylesheets = false;

	if (GetElementsWithAttrCount(dom, "link", "type", "*css*"))
		this->stylesheets = true;

	if (GetElementsWithAttrCount(dom, "style", "type", "*css*"))
		this->stylesheets = true;

	this->semantic_value = PSV_AVERAGE;
	if (!format_els)
		semantic_value = PSV_SEMANTIC_ONLY;
	else if ((format_els + semantic_els) / format_els > FORMATTED_ACCEPTABLE)
		semantic_value = PSV_FORMATTED;

	this->flash = false;
	if (GetElementsWithAttrCount(dom, "param", "value", "*.swf"))
		this->flash = true;
	if (!flash && GetElementsWithAttrCount(dom, "embed", "src", "*.swf"))
		this->flash = true;
}

ImageTypeCounts::ImageTypeCounts()
{
	this->gif_images = 0;
	this->jpg_images = 0;
	this->other_images = 0;
	this->png_images = 0;
}

ImageTypeCounts ImageTypeCounts::operator += (const ImageTypeCounts & c)
{
	this->gif_images += c.gif_images;
	this->jpg_images += c.jpg_images;
	this->other_images += c.other_images;
	this->png_images += c.png_images;

	return *this;
}

SiteStats::SiteStats()
{
	this->average_pages = 0;
	this->doctypes.clear();
	this->formatted_pages = 0;
	this->javascripted = 0;
	this->semantical_pages = 0;
	this->stylesheeted = 0;
	this->total_pages = 0;
	this->xml_declarations = 0;
	this->xmlfeeded = 0;
	this->withflash = 0;

#define ASD(y, x) this->serialize_map[(x)] = (y)
	ASD(&this->average_pages, "average");
	ASD(&this->formatted_pages, "formatted");
	ASD(&this->javascripted, "javascripted");
	ASD(&this->semantical_pages, "semantical");
	ASD(&this->stylesheeted, "stylesheeted");
	ASD(&this->total_pages, "totalpages");
	ASD(&this->xml_declarations, "xmldeclarations");
	ASD(&this->xmlfeeded, "xmlfeeded");
	ASD(&this->homepagelinked, "homepagelinked");
	ASD(&this->image_types.gif_images, "gifs");
	ASD(&this->image_types.jpg_images, "jpgs");
	ASD(&this->image_types.png_images, "pngs");
	ASD(&this->image_types.other_images, "otherimages");
#undef ASD
}

SiteStats SiteStats::operator += (const PageStats & p)
{
	this->doctypes[p.doctype]++;
	this->image_types += p.image_types;
	switch (p.semantic_value) {
	case PSV_FORMATTED: this->formatted_pages++; break;
	case PSV_AVERAGE: this->average_pages++; break;
	case PSV_SEMANTIC_ONLY: this->semantical_pages++; break;
	}
	
	this->total_pages++;

	if (p.xml_decl) this->xml_declarations++;
	if (p.stylesheets) this->stylesheeted++;
	if (p.javascript) this->javascripted++;
	if (p.xml_feed) this->xmlfeeded++;
	if (p.flash) this->withflash++;
	if (p.homepagelink) this->homepagelinked++;
	
	return *this;
}

ostream & operator << (ostream & o, const ImageTypeCounts & i)
{
	unsigned totali = i.jpg_images + i.png_images + i.gif_images + i.other_images;
	o << "<h3>Image formats</h3>" << endl;
	o << "<table>" << endl;

#define MYPRINT(format) o << \
	"<tr><td>" #format "</td><td>" << i.##format##_images << "</td><td>" << 100.0*i.##format##_images/totali << " %</td></tr>" << endl;

	o.setf(ios::fixed);
	MYPRINT(jpg);
	MYPRINT(png);
	MYPRINT(gif);
	MYPRINT(other);

#undef MYPRINT
	// o << "<tr><td>Other</td><td>" << i.other_images << "</td></tr>" << endl;
	o << "<tr><td>Total</td><td colspan=\"2\">" << totali << "</td></tr>" << endl;
	o << "</table>" << endl;

	return o;
}

ostream & operator << (ostream & o, const SiteStats & c)
{
	o << "<!DOCTYPE html>"
		"<html>"
		"<head><title>Web Analyzer Crawl Stats</title>"
		"<link rel='stylesheet' href='http://www.w3.org/StyleSheets/Core/Modernist' type='text/css'>"
		"<style type='text/css'><!--"
		"table td { padding: 0.2em; }"
		"</style>"
	    "</head>"
		"<body><h1>Crawl Stats</h1>" << endl <<
		"<dl>"
		"<dt>Total Pages</dt><dd>" << c.total_pages << "</dd>"
		"<dt>Pages with CSS</dt><dd>" << c.stylesheeted << "</dd>"
		"<dt>Pages with JavaScript</dt><dd>" << c.javascripted << "</dd>"
		"<dt>Pages with Flash</dt><dd>" << c.withflash << "</dd>"
		"<dt>Pages with RSS</dt><dd>" << c.xmlfeeded << "</dd>"
		"<dt>Pages with XML declaration</dt><dd>" << c.xml_declarations << "</dd>"
		"<dt>Pages with link to homepage</dt><dd>" << c.homepagelinked << "</dd>"
		"<dt>Non-semantic pages</dt><dd>" << c.formatted_pages << "</dd>"
		"<dt>Semi-semantic pages</dt><dd>" << c.average_pages << "</dd>"
		"<dt>Full-semantic pages</dt><dd>" << c.semantical_pages << "</dd>"
		"</dl>" << endl;
	
	map<string,unsigned int>::const_iterator e;

	o << "<h3>DOCTYPE declarations</h3>" << endl;
	o << "<table>"
		"<thead><tr><td>#</td><td>DOCTYPE</td></tr></thead>"
		"<tbody>" << endl;

	for(e = c.doctypes.begin(); e != c.doctypes.end(); e++)
	{
		o << "<tr><td>" << e->second << "</td><td>" << e->first << "</td></tr>" << endl;
	}

	o << "</tbody></table>" << endl;

	o << c.image_types << endl;

	o << endl << "</body></html>" << endl;
	return o;
}

std::istream & operator >> (istream & i, SiteStats & s)
{
	string s1;
	map<std::string, unsigned *>::iterator upp;
	unsigned count;

	while (!i.eof())
	{
		i >> s1;
		if (i.eof()) return i;

		if (!s1.compare("stat"))
		{
			i >> s1;
			if (i.eof()) return i;

			upp = s.serialize_map.find(s1);
			if (upp != s.serialize_map.end())
			{
				i >> *upp->second;
			}
		}
		else if (!s1.compare("doctype"))
		{
			char c;
			s1 = "";

			i >> count;
			if (i.eof()) return i;
			do {
				c = i.get();
			} while (!i.eof() && c != '<');
			
			if (i.eof()) return i;

			do {
				c = i.get();
				if (c != '>') s1 += c;
				else break;
			} while (!i.eof());

			s.doctypes[s1] = count;
		}
		else if (!s1.compare("statsend"))
		{
			return i;
		}
	}

	return i;
}

std::ostream & SiteStats::ProjectDump(std::ostream & o) const
{
	o << "statsstart" << endl;
	map<string, unsigned *>::const_iterator ci;

	for (ci = this->serialize_map.begin(); ci != serialize_map.end(); ci++)
	{
		o << "stat " << ci->first << " " << *ci->second << endl;
	}

	map<string, unsigned>::const_iterator di;
	for (di = doctypes.begin(); di != doctypes.end(); di++)
	{
		o << "doctype " << di->second << " <" << di->first << ">" << endl;
	}

	o << "statsend" << endl;

	return o;
}

}; // namespace
